;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; schematic/extract.scm
;;;
;;; See this project's README for more information.
;;;
;;; Copyright (c) 2013-2014, Evan Hanson <evhan@foldling.org>
;;; See LICENSE for details.
;;;

(cond-expand
  (chicken
   (import (only (chicken) symbol-append)))
  (else
   (define (symbol-append . symbols)
     (string->symbol
      (apply string-append (map symbol->string symbols))))))

;;;
;;; Read helpers.
;;;

;; `read`, but returns `#false` on error.
(define (read/guard port)
  (guard (_ (else #false))
    (read port)))

;;
;; Reads the `car` of the pair on `port`, returning `#false` if the read
;; operation signals an error or if no pair is encountered.
;;
;; Does not handle multiline or datum comments.
;;
(define (read-car port)
  ;; Discard whitespace and line comments.
  (do ((c (peek-char port) (peek-char port)))
      ((or (eof-object? c) (not (memq c '(#\; #\space #\tab #\newline)))))
    (case (read-char port)
      ((#\;) (read-line port))))
  ;; Read one datum past a single opening parenthesis.
  (case (read-char port)
    ((#\() (read/guard port))
    (else #false)))

;;;
;;; Basic pattern matching.
;;;

;; Matches `form` against `pattern`, returning a list of all elements
;; from `form` that corresponded to a vector in `pattern`, or `#false`
;; if the two forms didn't otherwise match (in the sense of `equal?`).
(define (match form pattern)
  (let ((env '()))
    (define (extract x p)
      (cond ((vector? p)
             (set! env (cons x env))
             #true)
            ((and (pair? x) (pair? p))
             (and (extract (car x) (car p))
                  (extract (cdr x) (cdr p))))
            (else (equal? x p))))
    (and (extract form pattern)
         (reverse env))))

;; `case`-like syntax for `match`, above.
(define-syntax match-case
  (syntax-rules (else =>)
    ((_ form) (if #false #true))
    ((_ form (else . body)) (begin . body))
    ((_ form (pattern => proc) . clauses)
     (let ((results (match form (quote pattern))))
       (if (not results)
           (match-case form . clauses)
           (apply proc results))))
    ((_ form (pattern . body) . clauses)
     (if (match form (quote pattern))
         (begin . body)
         (match-case form . clauses)))))

;;;
;;; Output.
;;;

(define emit display)

(define (emit-documentation docs)
  (write docs))

(define (emit-specification type form)
  (write (cons type form)))

;;;
;;; Definition extraction.
;;;

;;
;; Extracts commented definitions from the Scheme code on `input`,
;; writing s-expressive specifications for each to `output`.
;;
;; Each of these specifications is of the form:
;;
;;     <specification> = (<comment> (<type> . <form>))
;;     <comment>       = string?
;;     <form>          = any?
;;     <type>          = 'procedure | 'syntax | 'constant
;;                     | 'parameter | 'record | 'string
;;
(define extract-definitions
  (case-lambda
    (()
     (extract-definitions '(";;" ";;;") (current-input-port) (current-output-port)))
    ((input)
     (extract-definitions '(";;" ";;;") input (current-output-port)))
    ((input output)
     (extract-definitions '(";;" ";;;") input output))
    ((comment-prefixes input output)
     (parameterize ((current-output-port output))
       (port-fold-source-sections
        (lambda (docs code carry)
          (when (pair? carry)
            (set! docs (car carry))
            (set! code (string-append (cdr carry) code)))
          (cond
            ((string=? docs ""))
            ((call-with-port (open-input-string code) read/guard) =>
             (lambda (form)
               (match-case form
                 ((define (#(name) . #(args)) . #(_)) =>
                  (lambda (name args _)
                    (emit #\()
                    (emit-documentation docs)
                    (emit-specification 'procedure (cons name args))
                    (emit #\))))
                 ((define #(name) (lambda #(args) . #(_))) =>
                  (lambda (name args _)
                    (emit #\()
                    (emit-documentation docs)
                    (emit-specification 'procedure (cons name args))
                    (emit #\))))
                 ((define #(name) (case-lambda . #(clauses))) =>
                  (lambda (name clauses)
                    (emit #\()
                    (emit-documentation docs)
                    (for-each (lambda (args)
                                (emit-specification 'procedure (cons name args)))
                              (map car clauses))
                    (emit #\))))
                 ((define #(name) (make-parameter . #(_))) =>
                  (lambda (name _)
                    (emit #\()
                    (emit-documentation docs)
                    (emit-specification 'parameter name)
                    (emit #\))))
                 ((define #(name) #(value)) =>
                  (lambda (name value)
                    (cond ((string? value)
                           (emit #\()
                           (emit-documentation docs)
                           (emit-specification 'string name)
                           (emit #\)))
                          ((not (pair? value))
                           (emit #\()
                           (emit-documentation docs)
                           (emit-specification 'constant name)
                           (emit #\))))))
                 ((define-syntax #(name) (syntax-rules #(_) . #(clauses))) =>
                  (lambda (name _ clauses)
                    (emit #\()
                    (emit-documentation docs)
                    (for-each (lambda (args)
                                (emit-specification 'syntax (cons name args)))
                              (map cdar clauses))
                    (emit #\))))
                 ((define-syntax #(name) . #(_)) =>
                  (lambda (name _)
                    (emit #\()
                    (emit-documentation docs)
                    (emit-specification 'syntax name)
                    (emit #\))))
                 ((define-record-type #(name) #(make) #(pred) . #(fields)) =>
                  (lambda (name make pred fields)
                    (emit #\()
                    (emit-documentation docs)
                    (emit-specification 'record name)
                    (emit-specification 'procedure make)
                    (emit-specification 'procedure `(,pred object))
                    (for-each (lambda (p)
                                (emit-specification 'procedure `(,p ,name)))
                              (map cadr fields))
                    (for-each (lambda (p)
                                (when (pair? p)
                                  (emit-specification 'procedure `(,(car p) ,name value))))
                              (map cddr fields))
                    (emit #\))))
                 ((define-record #(name) . #(fields)) => ; CHICKEN shorthand.
                  (lambda (name fields)
                    (emit #\()
                    (emit-documentation docs)
                    (emit-specification 'record name)
                    (emit-specification 'procedure `(,(symbol-append 'make- name) ,@fields))
                    (emit-specification 'procedure `(,(symbol-append name '?) object))
                    (for-each (lambda (f)
                                (emit-specification 'procedure `(,(symbol-append name '- f) ,name)))
                              fields)
                    (for-each (lambda (f)
                                (emit-specification 'procedure `(,(symbol-append name '- f '-set!) ,name value)))
                              fields)
                    (emit #\)))))))
            (else
             (case (call-with-port (open-input-string code) read-car)
               ((define-library) #false)   ; R7RS.
               ((module) #false)           ; CHICKEN.
               ((define-module) #false)    ; Gauche.
               (else (cons docs code)))))) ; Carry into the next section.
        #false
        comment-prefixes
        input)))))
